import pprint


class BaseDataFormatter:
    def __init__(self, non_verbose_fields: list[list | str] = []):
        self.non_verbose_fields = non_verbose_fields

    def obj_to_dict(self, obj):
        if isinstance(obj, (int, float, str, bool)):
            return obj 
        elif isinstance(obj, dict):
            return {k: self.obj_to_dict(v) for k, v in obj.items()} 
        elif isinstance(obj, (list, tuple, set)):
            return type(obj)(self.obj_to_dict(item) for item in obj) 
        elif obj == None:
            return None
        elif isinstance(obj, object):
            return self.obj_to_dict(obj.__dict__) 
        return None 
    
    def non_verbose_formatter(self, serialized_data):
        formatted = dict()
        if self.non_verbose_fields:
            for field in self.non_verbose_fields:
                if isinstance(field, str):
                    formatted[field] = serialized_data[field]
                elif isinstance(field, list):
                    formatted[field] = self.non_verbose_formatter(serialized_data[field])
            return formatted
        else:
            return serialized_data
    
    def __call__(self, data, verbose):
        serialized = self.obj_to_dict(data)
        if not isinstance(data, (str, int)):
            if verbose:
                return serialized
            if isinstance(serialized, list) and self.non_verbose_fields:
                return_data = list()
                for fragment in serialized:
                    non_verbose_fragment = self.non_verbose_formatter(fragment)
                    return_data.append(non_verbose_fragment)
                return return_data
            return self.non_verbose_formatter(serialized)
        else:
            return data
        

    



'''
SCHEMAS
These are standardized JSON serializable data formats to make socket operations on the localhost client-server smoother.
'''
from typing import Any, Optional, Literal

from pydantic import BaseModel, Field, validator


class BaseSchema(BaseModel):
    schema_type: str
    request_content: dict = dict()
    message: dict = dict()
    error: str = str()
    system: str = str()
    pwd: str = str()


class CommandData(BaseSchema):
    """
    Command execution request sent from the client to the server to execute a specified command
    """
    schema_type = 'CommandData'


class PasswordAuthData(BaseSchema):
    """
    Sends credentials from the client to the server for auth to Tapis services
    """
    schema_type: str = 'PasswordAuthData'
    username: Optional[str]
    password: Optional[str]


class StartupData(BaseSchema):
    """
    used only during connection inintialization between the client and server to transmit startup data
    """
    schema_type: str = 'StartupData'
    initial: bool = False
    username: Optional[str]
    url: Optional[str]


class ResponseData(BaseSchema):
    """
    data from the server to the client with return data from commands, as well as errors
    """
    schema_type: str = 'ResponseData'
    url: str = str()
    active_username: str = str()
    exit_status = 0
    

class FormRequest(BaseSchema):
    """
    Request seperate input for some command parameters. If the arguments_list is empty, this will be interpreted as an expression request for something like neo4j
    """
    schema_type: str = 'FormRequest'
    existing_data: dict | object = dict()


class FormResponse(BaseSchema):
    """
    respond to a form request with proper data
    """
    schema_type: str = 'FormResponse'


class AuthRequest(BaseSchema):
    """
    Request auth credentials from the client
    """
    schema_type: str = 'AuthRequest'
    auth_request_type: Literal["password", "device_code", "federated", "requested", "success"]


class ConfirmationRequest(BaseSchema):
    """
    ask the client for confirmation to carry out an action
    """
    schema_type: str = 'ConfirmationRequest'


"""
Requirements:
Add a way to set a default system to be accessed. You must be able to override this by explicitly specifying system id in command
add intuitive file access and manipulation using this default system. 
"""


import json
import os
import pprint
import webbrowser


if __name__ != "__main__":
    from .. import baseCommand, decorators
    from ..arguments import argument
    from .. import commandOpts
    from socketopts import schemas
    from . import systemForms
    Argument = argument.Argument


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
base_config_path = os.path.join(__location__, '..\\..\\..\\tapis-config-files\\system-config.json')


class system(baseCommand.BaseCommand):
    """
    @help: set the system to run operations on by default. Running this will put you "in" the system so that you dont have to specify system ID for each command
    """
    required_arguments=[
        Argument('systemId', size_limit=(1, 80), positional=True),
    ]
    async def run(self, *args, **kwargs):
        system_info = self.t.systems.getSystem(systemId=kwargs['systemId'])
        kwargs['connection'].system = kwargs['systemId']
        kwargs['connection'].pwd = system_info.rootDir
        return f"successfully entered the system {kwargs['connection'].system}"


class exit_system(baseCommand.BaseCommand):
    """
    @help: exit the default system
    """
    async def run(self, *args, **kwargs):
        kwargs['connection'].system = ''
        kwargs['connection'].pwd = ''
        return "successfully exited system"
    

class get_systems(baseCommand.BaseCommand):
    """
    @help: Gets and returns the list of systems the current Tapis service and account have access to
    @doc: this is an example of the doc segment of the docstring. not included in help message
    """
    return_fields = ['id', 'systemType', 'host']
    optional_arguments = [
        Argument('listType', choices=['OWNED', 'SHARED_PUBLIC', 'ALL'])
    ]
    async def run(self, *args, **kwargs):
        pprint.pprint(kwargs)
        systems = self.t.systems.getSystems(**kwargs)
        return systems
    

class get_system(baseCommand.BaseCommand):
    """
    @help: get information on a selected system
    """
    return_fields = ['id', 'systemType', 'host', 'enabled']
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments=[
        Argument('systemId', size_limit=(1, 80), positional=True)
    ]
    async def run(self, *args, **kwargs): 
        system_info = self.t.systems.getSystem(systemId=kwargs['systemId'])
        return system_info
    

class verify_pki_keys(baseCommand.BaseCommand):
    """
    @help: upload system credentials to a system. Must generate keys first using 'ssh-keygen -m PEM -f id_rsa', and format with, 'awk -v ORS='\\n' '1' <private_key_name>
    file argument must contain the path to the private and public keys respectively, separated by a ','
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('systemId', positional=True),
        Argument('private_key_path'),
        Argument('public_key_path')
    ]
    async def run(self, *args, **kwargs): # get information about a system given its ID
        with open(kwargs['private_key_path'], 'r') as f:
            private_key = f.read()

        with open(kwargs['public_key_path'], 'r') as f:
            public_key = f.read()

        cred_return_value = self.t.systems.createUserCredential(systemId=kwargs['systemId'],
                            userName=self.username,
                            privateKey=private_key,
                            publicKey=public_key)

        return str(cred_return_value)
    

class get_scheduler_profiles(baseCommand.BaseCommand):
    """
    @help: get list of scheduler profiles in order to use while creating system
    """
    async def run(self, *args, **kwargs):
        return [{"name":scheduler.name, "description":scheduler.description, "tenant":scheduler.tenant} for scheduler in self.t.systems.getSchedulerProfiles()]
    

class create_scheduler_profile(baseCommand.BaseCommand):
    """
    @help: create a new Tapis scheduler profile with which to run batch jobs on a system
    """
    required_arguments = [
        Argument('name')
    ]
    optional_arguments = [
        Argument('description', arg_type='str_input'),
        Argument('owner', default_value=r"${apiUserId}"),
        Argument('moduleLoads', arg_type='input_list',
                  data_type=argument.Form('moduleLoad',
                    required_arguments = [
                        Argument('moduleLoadCommand')
                    ],
                    optional_arguments = [
                        Argument('modulesToLoad', arg_type='input_list', data_type=Argument('module'))
                    ]))
    ]
    async def run(self, *args, **kwargs):
        return self.t.systems.createSchedulerProfile(**kwargs)


class delete_scheduler_profile(baseCommand.BaseCommand):
    """
    @help: delete a scheduler profile
    """
    required_arguments = [
        Argument('name'),
        Argument('confirm', arg_type='confirmation')
    ]
    async def run(self, *args, **kwargs):
        return self.t.systems.deleteSchedulerProfile(**kwargs)
    

class SystemAuth:
    sys_auth_map = None
    def config_auth_map(self):
        self.sys_auth_map = {"LINUX":{"PASSWORD":self.password_auth, "PKI_KEYS":self.pki_keys_auth}, 
                             "S3":{"ACCESS_KEY":self.access_key_auth}, "GLOBUS":{"TOKEN":self.token_auth}, 
                             "IRODS":{"TOKEN":self.password_auth}}
        
    async def password_auth(self, **kwargs):
        request = schemas.FormRequest(request_content={'username':Argument('username', arg_type='str_input'), "password":Argument('password', arg_type='secure')},
                                        message={'message':f"Enter your credentials to the select host specified in the system creation"})
        await kwargs['connection'].send(request)
        response = await kwargs['connection'].receive()
        response_content = response.request_content
        password = response_content['password']
        loginUserName = response_content['username']
        cred_return_value = self.t.systems.createUserCredential(systemId=kwargs['id'],
                            userName=self.username,
                            loginUser=loginUserName,
                            password=password)
        return cred_return_value
    
    async def token_auth(self, **kwargs):
        session_details = self.systems.getGlobusAuthUrl()
        request = schemas.FormRequest(message={"message":f"travel to the url {session_details.url} if the page doesnt open on its own, and enter your credentials. Then paste the code displayed to this app"},
                                      request_content={'Auth_code':Argument('code', arg_type='str_input')})
        webbrowser.open(session_details.url)
        await kwargs['connection'].send(request)
        response = await kwargs['connection'].receive()
        auth_code = response.request_content['Auth_code']
        token_info = self.t.systems.generateGlobusTokens(systemId=kwargs['id'], userName=self.username, authCode=auth_code, sessionId=session_details.sessionId)
        cred_return_value = self.t.systems.createUserCredential(systemId=kwargs['id'],
                                                                userName=self.username,
                            accessToken=token_info.access_token,
                            refreshToken=token_info.refresh_token)
        return cred_return_value
    
    async def access_key_auth(self, **kwargs):
        request = schemas.FormRequest(message={"message":f"Retrieve the access key and access secret from your S3 bucket"},
                                      request_content={'access_key':Argument('access_key', arg_type='str_input'), 'access_secret':Argument('access_secret', arg_type='secure')})
        await kwargs['connection'].send(request)
        response = await kwargs['connection'].receive()
        cred_return_value = self.t.systems.createUserCredential(systemId=kwargs['id'],
                                                                userName=self.username,
                            accessKey=response.request_content['access_key'],
                            password=response.request_content['access_secret'])
        return cred_return_value
    
    async def pki_keys_auth(self, **kwargs):
        request = schemas.FormRequest(message= {"message":
                                      f"""Follow these steps to manually register PKI keys with the system
                                      1. ssh into the host you are creating your system on, {kwargs['host']}
                                      2. navigate into the .ssh folder of the host. If there are not pre-generated keys, run ssh-keygen -t rsa -b 4096 -m PEM
                                      3. format both key files using cat $privateKeyFile | awk -v ORS='\\n' '1'. Both keys must be in single line format
                                      4. ensure the private key is not in openSSH format (that is it has the begin and end headers)
                                      5. download both to your local machine using scp
                                      6. run the verify_pki_keys command and submit the system id {kwargs['id']} and the file path of the public and private keys
                                      7. you should be able to access the system now"""},
                                      request_content={"continue":Argument("continue", arg_type='confirmation')})
        await kwargs['connection'].send(request)
        await kwargs['connection'].receive()
        return None        
    
    async def base_system_auth(self, kwargs):
        if kwargs['defaultAuthnMethod'] not in self.sys_auth_map[kwargs['systemType']]:
            raise ValueError(f"The system type {kwargs['systemType']} does not support {kwargs['defaultAuthnMethod']} authentication.")
        else:
            try:
                auth_result = await self.sys_auth_map[kwargs['systemType']][kwargs['defaultAuthnMethod']](**kwargs)
                return auth_result
            except Exception as e:
                raise Exception(f"System creation for {kwargs['id']} succeeded, but Authentication for the system failed, maybe you entered something wrong? run the submit_system_credentials command")
    
    async def authenticate(self, kwargs):
        return await self.sys_auth_map[kwargs['systemType']][kwargs['defaultAuthnMethod']](**kwargs)


class create_system(baseCommand.BaseCommand, SystemAuth):
    """
    @help: create a system. Must have a properly configured system file.
    see the template at https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/blob/main/icicle_rel_04_2023/CLI/TapisCL-ICICLE/tapis-config-files/system-config.json
    this command will automatically create and upload the ssh keys
    """
    supports_config_file=True
    required_arguments=[
        Argument('id', size_limit=(1, 80), positional=True),
        Argument('systemType', choices=["LINUX", "S3", "IRODS", "GLOBUS"], description=
                                    """LINUX is a standard linux system
IRODS refers to an IRODS data management system
GLOBUS refers to a GLOBUS file system"""),
        Argument('host', size_limit=(1, 256), description="In the case of Linux this is the hostname or IP of the HPC system you want to connect to. In the case of Globus, this is a Globus ID"),
        Argument('defaultAuthnMethod', choices=['PASSWORD', "PKI_KEYS", "TOKEN"], description=
                                    """Depending on your systemType, you will be restricted to certain options.
LINUX: PASSWORD, PKI_KEYS
GLOBUS: TOKEN
IRODS: PASSWORD
In the case you choose password, your username and password will either be your TACC account info, or the login info you used with federated/device_code grant"""),
    ]
    optional_arguments=[
        systemForms.CAN_EXEC,
        systemForms.USE_PROXY,
        systemForms.CAN_RUN_BATCH,
        systemForms.MOUNT_DATA_TRANSFER_NODE,
        systemForms.JOB_CONSTRAINTS,
        systemForms.JOB_CAPABILITIES,
        systemForms.JOB_ENVIRONMENT_VARIABLES,
        Argument('effectiveUserId', default_value=r"${apiUserId}", size_limit=(0, 60)),
        Argument('owner', default_value=r"${apiUserId}"),
        Argument('description', arg_type='str_input', size_limit=(0, 2048)),
        Argument('enabled', action='store_true'),
        Argument('rootDir', default_value='/', size_limit=(0, 4096)),
        Argument('port', data_type='int', default_value=22),
        Argument('isDtn', action='store_true', depends_on=['rootDir'], mutually_exclusive_with=['mountDataTransferNode', 'canExec']),
        Argument('enableCmdPrefix', action='store_true'),
        Argument('mpiCmd', size_limit=(0, 126), arg_type='str_input'),
        Argument('tags', arg_type='input_list', data_type=Argument('tag', arg_type='str_input')),
        Argument('notes', arg_type='str_input'),
        Argument('importRefId')
    ]
    def __init__(self):
        super().__init__()
        self.config_auth_map()

    async def run(self, *args, **kwargs) -> str: # create a tapius system. Takes a path to a json file with all system information, as well as an ID
        return_info = dict()
        return_info['system_creation_info'] = self.t.systems.createSystem(**kwargs)
        return_info['auth_result'] = await self.base_system_auth(kwargs)
        return return_info
    

class create_s3_system(create_system):
    """
    @help: create a system specially configured to support AWS s3 buckets
    """
    required_arguments = [
        Argument('id', size_limit=(1, 80), positional=True),
        Argument('host', size_limit=(1, 256), description="AWS S3 bucket url"),
        Argument('bucketName')
    ]
    optional_arguments = [
        systemForms.USE_PROXY,
        systemForms.CAN_RUN_BATCH,
        systemForms.MOUNT_DATA_TRANSFER_NODE,
        Argument('effectiveUserId', default_value=r"${apiUserId}", size_limit=(0, 60)),
        Argument('owner', default_value=r"${apiUserId}"),
        Argument('description', arg_type='str_input', size_limit=(0, 2048)),
        Argument('enabled', action='store_true'),
        Argument('rootDir', default_value='/', size_limit=(0, 4096)),
        Argument('port', data_type='int', default_value=22),
        Argument('enableCmdPrefix', action='store_true'),
        Argument('mpiCmd', size_limit=(0, 126), arg_type='str_input'),
        Argument('tags', arg_type='input_list', data_type=Argument('tag', arg_type='str_input')),
        Argument('notes', arg_type='str_input'),
        Argument('importRefId')
    ]
    async def run(self, *args, **kwargs):
        kwargs['defaultAuthnMethod'] = 'ACCESS_KEY'
        kwargs['systemType'] = 'S3', 
        return_info = dict()
        return_info['system_creation_info'] = self.t.systems.createSystem(**kwargs)
        return_info['auth_result'] = await self.base_system_auth(kwargs)
        return return_info


class submit_system_credentials(create_system):
    """
    @help: manually submit system credentials
    """
    supports_config_file = False
    required_arguments = [
        Argument('id', positional=True)
    ]
    optional_arguments = []
    async def run(self, *args, **kwargs):
        system_info = self.t.systems.getSystem(systemId=kwargs['id'])
        kwargs['systemType'] = system_info.systemType
        kwargs['defaultAuthnMethod'] = system_info.defaultAuthnMethod
        kwargs['host'] = system_info.host
        return_info = await self.authenticate(kwargs)
        return return_info
    

class SystemUpdatingRetriever(baseCommand.UpdatableFormRetriever):
    def __call__(self, tapis_instance, **kwargs):
        system_data = tapis_instance.systems.getSystem(systemId=kwargs["systemId"])
        return system_data
    

class update_system(baseCommand.BaseCommand):
    """
    @help: update a system with new information
    """
    updateable_form_retriever = SystemUpdatingRetriever()
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments=[
        Argument('systemId', size_limit=(1, 80), positional=True),
    ]
    optional_arguments=[
        systemForms.USE_PROXY,
        systemForms.CAN_RUN_BATCH,
        systemForms.MOUNT_DATA_TRANSFER_NODE,
        systemForms.JOB_CONSTRAINTS,
        systemForms.JOB_CAPABILITIES,
        systemForms.JOB_ENVIRONMENT_VARIABLES,
        Argument('defaultAuthnMethod', choices=['PASSWORD', "PKI_KEYS", "TOKEN"], description=
                                    """Depending on your systemType, you will be restricted to certain options.
LINUX: PASSWORD, PKI_KEYS
GLOBUS: TOKEN
IRODS: PASSWORD
In the case you choose password, your username and password will either be your TACC account info, or the login info you used with federated/device_code grant"""),
        Argument('jobRuntimes', arg_type='input_list', data_type=argument.Form(
            'jobRuntime', required_arguments = [
                    Argument('runtimeType', choices=['DOCKER', 'SINGULARITY']), 
                ],
                optional_arguments=[
                    Argument('version')
                ]
            )),
        Argument('jobWorkingDir', default_value=r"HOST_EVAL($WORK2)", size_limit=(0, 4096), description='Where on this hpc system are jobs run?'),
        Argument('effectiveUserId', default_value=r"${apiUserId}", size_limit=(0, 60)),
        Argument('description', arg_type='str_input', size_limit=(0, 2048)),
        Argument('port', data_type='int', default_value=22),
        Argument('enableCmdPrefix', action='store_true'),
        Argument('mpiCmd', size_limit=(0, 126), arg_type='str_input'),
        Argument('tags', arg_type='input_list', data_type=Argument('tag', arg_type='str_input')),
        Argument('notes', arg_type='str_input'),
        Argument('importRefId')
    ]
    async def run(self, *args, **kwargs):
        result = self.t.systems.patchSystem(**kwargs)
        return result
    

class update_s3_system(update_system):
    """
    @help: update an s3 system
    """
    optional_arguments = [
        systemForms.USE_PROXY,
        systemForms.CAN_RUN_BATCH,
        systemForms.MOUNT_DATA_TRANSFER_NODE,
        Argument('effectiveUserId', default_value=r"${apiUserId}", size_limit=(0, 60)),
        Argument('description', arg_type='str_input', size_limit=(0, 2048)),
        Argument('port', data_type='int', default_value=22),
        Argument('enableCmdPrefix', action='store_true'),
        Argument('mpiCmd', size_limit=(0, 126), arg_type='str_input'),
        Argument('tags', arg_type='input_list', data_type=Argument('tag', arg_type='str_input')),
        Argument('notes', arg_type='str_input'),
        Argument('importRefId')
    ]
    async def run(self, *args, **kwargs):
        result = self.t.systems.patchSystem(**kwargs)
        return result


class is_system_enabled(baseCommand.BaseCommand):
    """
    @help: check to see if a system is enabled
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('systemId', positional=True)
    ]
    async def run(self, *args, **kwargs):
        return self.t.systems.isEnabled(**kwargs)


class enable_system(is_system_enabled):
    """
    @help: enable a system
    """
    async def run(self, *args, **kwargs):
        return self.t.systems.enableSystem(**kwargs)
    

class disable_system(is_system_enabled):
    """
    @help: disable a system
    """
    async def run(self, *args, **kwargs):
        return self.t.systems.disableSystem(**kwargs)
    

class delete_system(baseCommand.BaseCommand):
    """
    @help: delete the selected system
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments=[
        Argument('systemId', size_limit=(1, 80), positional=True),
        Argument('confirm', arg_type='confirmation')
    ]
    async def run(self, *args, **kwargs) -> str:
        kwargs['connection'].system = ''
        kwargs['connection'].pwd = ''
        return_value = self.t.systems.deleteSystem(systemId=kwargs['systemId'])
        return return_value
    

class undelete_system(baseCommand.BaseCommand):
    """
    @help: undo deletion
    """
    required_arguments=[
        Argument('systemId', size_limit=(1, 80), positional=True),
    ]
    async def run(self, *args, **kwargs):
        return self.t.systems.undeleteSystem(**kwargs)

class create_child_system(baseCommand.BaseCommand):
    """
    @help: create a child system which inherits majority attributes from parent
    """
    required_arguments = [
        Argument('parentId', positional=True),
        Argument('rootDir')
    ]
    optional_arguments = [
        Argument('effectiveUserId', default_value=r"${apiUserId}")
    ]
    async def run(self, *args, **kwargs):
        return self.t.systems.createChildSystem(**kwargs)
    

class unlink_child_system(baseCommand.BaseCommand):
    """
    @help: make a child system stand alone, so that updates to parent system will not transfer to this child system
    """
    required_arguments = [
        Argument('parentId', positional=True)
    ]
    async def run(self, *args, **kwargs):
        return self.t.systems.unlinkFromParent(**kwargs)
    

class unlink_children(baseCommand.BaseCommand):
    """
    @help: unlink an array of child systems from their parent system. Requires parent system permissions
    """
    required_arguments = [
        Argument('parentId', positional=True),
        Argument('childSystemIds', arg_type='input_list', data_type=Argument('systemId', size_limit=(1, 80)))
    ]
    async def run(self, *args, **kwargs):
        parent_id = kwargs.pop('parentId')
        kwargs['parentSystemId'] = parent_id
        return self.t.systems.unlinkChildren(**kwargs)
    

class get_user_perms(baseCommand.BaseCommand):
    """
    @help: list perms for the user on a select system
    """
    required_arguments = [
        Argument('systemId'),
        Argument('userName')
    ]
    async def run(self, *args, **kwargs):
        return self.t.systems.getUserPerms(**kwargs)
    

class grant_user_perms(baseCommand.BaseCommand):
    """
    @help: assign a perm for a user on the select system
    """
    required_arguments = [
        Argument('systemId'),
        Argument('userName'),
        Argument('permissions', arg_type='input_list', data_type=Argument('permission', choices=['READ', 'MODIFY', 'EXECUTE'], arg_type='str_input'))
    ]
    async def run(self, *args, **kwargs):
        return self.t.grantUserPerms(**kwargs)
    

class revoke_user_perms(grant_user_perms):
    """
    @help: delete user permissions
    """
    async def run(self, *args, **kwargs):
        return self.t.revokeUserPerms(**kwargs)
    




import psycopg2


if __name__ != "__main__":
    from .. import baseCommand, decorators
    from ..arguments import argument


class postgres(baseCommand.BaseQuery):
    """
    @help: integrated CLI to interface with Postgres pods
    """
    required_arguments = [
        argument.Argument('id'),
        argument.Argument('expression', arg_type='expression')
    ]
    async def run(self, *args, **kwargs) -> str:
        uname, pword = self.get_pod_credentials(kwargs['id'])
        with psycopg2.connect(f"postgresql://{uname}:{pword}@{kwargs['id']}.pods.{self.t.base_url.split('https://')[1]}:443") as conn:
            conn.autocommit = True
            with conn.cursor() as cur:
                cur.execute(query=kwargs['expression'])
                return_value = cur.fetchall()
        return str(f'[+][{kwargs["id"]}] {return_value}')


from tapipy.tapis import errors as TapisErrors


if __name__ != "__main__":
    from . import baseCommand, decorators
    from .arguments import argument
    from . import commandOpts
    Argument = argument.Argument


class get_volumes(baseCommand.BaseCommand):
    """
    @help: get a list of volumes registered to your account
    """
    async def run(self, *args, **kwargs):
        return self.t.pods.get_volumes()
    

class create_volume(baseCommand.BaseCommand):
    """
    @help: create a new volume on which to store files to be used by Tapis pods. This allows several pods to share the same persistent files
    """
    supports_config_file=True
    required_arguments = [
        Argument('volume_id', positional=True)
    ]
    optional_arguments = [
        Argument('description', arg_type='str_input'),
        Argument('size_limit', data_type='int', default_value=1024)
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.create_volume(**kwargs)
    

class get_volume(baseCommand.BaseCommand):
    """
    @help: get information on a specific volume
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('volume_id')]
    required_arguments = [
        Argument('volume_id', positional=True)
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.get_volume(**kwargs)
    

class volume(baseCommand.BaseCommand):
    """
    @help: enter the volume to interact with its files more directly
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('volume_id')]
    required_arguments = [
        Argument('volume_id', positional=True)
    ]
    async def run(self, *args, **kwargs):
        volume_info = self.t.pods.get_volume(volume_id=kwargs['volume_id'])
        kwargs['connection'].system = volume_info.volume_id
        kwargs['connection'].pwd = "/"
        return f"successfully entered the volume {kwargs['connection'].system}"
    

class exit_volume(baseCommand.BaseCommand):
    """
    @help: exit the volume you are currently in 
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('volume_id')]
    required_arguments = [
        Argument('volume_id', positional=True)
    ]
    async def run(self, *args, **kwargs):
        kwargs['connection'].system = ""
        kwargs['connection'].pwd = ""
        return f"successfully exited the volume {kwargs['volume_id']}"
    

class update_volume(create_volume):
    """
    @help: update a volume's information
    """
    supports_config_file=True
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('volume_id')]
    async def run(self, *args, **kwargs):
        return self.t.pods.update_volume(**kwargs)
    

class delete_volume(baseCommand.BaseCommand):
    """
    @help: delete a volume
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('volume_id')]
    required_arguments = [
        Argument('volume_id', positional=True),
        Argument('confirm', arg_type='confirmation')
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.delete_volume_files(**kwargs)


class dir(baseCommand.BaseCommand):
    """
    @help: list all files in the selected volume
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('volume_id')]
    required_arguments = [
        Argument('volume_id', positional=True)
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.list_volume_files(**kwargs)
    

class upload_volume(baseCommand.BaseCommand):
    """
    @help: upload a file from your local machine to the volume
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('volume_id'), commandOpts.CHECK_PWD('path',)]
    required_arguments = [
        Argument('volume_id', positional=True),
        Argument('path'),
        Argument('file')
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.upload_to_volume(**kwargs)
    

class set_volume_permission(baseCommand.BaseCommand):
    """
    @help: set the permission for a user on a volume
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('volume_id')]
    required_arguments = [
        Argument('volume_id', positional=True),
        Argument('username'),
        Argument('level')
    ]
    async def run(self, *args, **kwargs):
        user = kwargs.pop('username')
        kwargs['user'] = user
        return self.t.pods.set_volume_permission(**kwargs)


class get_volume_permissions(baseCommand.BaseCommand):
    """
    @help: set the permission for a user on a volume
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('volume_id')]
    required_arguments = [
        Argument('volume_id', positional=True),
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.get_volume_permission(**kwargs)
    

class delete_volume_permission(baseCommand.BaseCommand):
    """
    @help: delete the persmission for a user on a volume
    """
    required_arguments = [
        Argument('volume_id', positional=True),
        Argument('username'),
        Argument('level')
    ]
    async def run(self, *args, **kwargs):
        user = kwargs.pop('username')
        kwargs['user'] = user
        return self.t.pods.delete_volume_permission(**kwargs)
    

class get_snapshots(baseCommand.BaseCommand):
    """
    @help: get a list of snapshots on the tenant you have access to
    """
    async def run(self, *args, **kwargs):
        return self.t.pods.get_snapshots()
    

class create_snapshot(baseCommand.BaseCommand):
    """
    @help: create a backup of the volume selected or specific files in the volume
    """
    supports_config_file=True
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('source_volume_id'), commandOpts.CHECK_PWD(('source_volume_path', 'destination_path'))]
    required_arguments = [
        Argument('snapshot_id', positional=True),
        Argument('source_volume_id', positional=True),
        Argument('source_volume_path')
    ]
    optional_arguments = [
        Argument('destination_path', description="Required if your snapshot is of a single file"),
        Argument('description', arg_type='str_input'),
        Argument('size_limit', data_type='int', description='maximum size in megabytes of the snapshot'),
        Argument('cron'),
        Argument('retention_policy')
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.create_snapshot(**kwargs)
    

class get_snapshot(baseCommand.BaseCommand):
    """
    @help: get snapshot information
    """
    required_arguments = [
        Argument('snapshot_id', positional=True)
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.get_snapshot(**kwargs)
    

class update_snapshot(baseCommand.BaseCommand):
    """
    @help: update snapshot information
    """
    supports_config_file=True
    required_arguments = [
        Argument('snapshot_id', positional=True)
    ]
    optional_arguments = [
        Argument('description', arg_type='str_input'),
        Argument('size_limit', data_type='int'),
        Argument('cron'),
        Argument('retention_policy')
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.update_snapshot(**kwargs)
    

class delete_snapshot(baseCommand.BaseCommand):
    """
    @help: delete snapshot information
    """
    required_arguments = [
        Argument('snapshot_id', positional=True),
        Argument('confirm', arg_type='confirmation')
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.delete_snapshot(**kwargs)
    

class list_snapshot_files(baseCommand.BaseCommand):
    """
    @help: list the files in a snapshot
    """
    required_arguments = [
        Argument('snapshot_id', positional=True)
    ]
    async def run(self, *args, **kwargs):
        return self.t.pods.list_snapshot_files(**kwargs)
    


from py2neo import Graph


if __name__ != "__main__":
    from .. import baseCommand, decorators
    from ..arguments import argument


class neo4j(baseCommand.BaseQuery):
    """
    @help: integrated CLI to interface with Neo4j pods
    """
    required_arguments = [
        argument.Argument('id'),
        argument.Argument('expression', arg_type='expression')
    ]
    async def run(self, *args, **kwargs) -> str: # function to submit queries to a Neo4j knowledge graph
        uname, pword = self.get_pod_credentials(kwargs["id"])
        graph = Graph(f"bolt+ssc://{kwargs['id']}.pods.{self.t.base_url.split('https://')[1]}:443", auth=(uname, pword), secure=True, verify=True)

        return_value = graph.run(kwargs['expression'])
        print(type(return_value))
        if str(return_value) == '(No data)' and 'create' in kwargs['expression'].lower(): # if no data is returned (mostly if something is created) then just say 'success'
            return f'[+][{kwargs["id"]}@pods.{self.t.base_url.split("https://")[1]}:443] Success'
        elif str(return_value) == '(No data)':
            return f'[-][{kwargs["id"]}@pods.{self.t.base_url.split("https://")[1]}:443] KG is empty'

        return return_value
    

class get_neo4j_pod_uri(baseCommand.BaseCommand):
    """
    @help: get the uri for a neo4j pod to use in external database software
    """
    required_arguments = [
        argument.Argument('id')
    ]
    async def run(self, *args, **kwargs) -> str:
        return f"bolt+ssc://{kwargs['id']}.pods.{self.t.base_url.split('https://')[1]}:443"


import json
import typing
from pprint import pformat

import pydantic


try:
    from socketopts import schemas
    from utilities import logger, exceptions
except:
    from . import schemas
    from ..utilities import logger, exceptions


schema_types: dict = {
        'CommandData':schemas.CommandData,
        'StartupData':schemas.StartupData,
        'ResponseData':schemas.ResponseData,
        'FormRequest':schemas.FormRequest,
        'FormResponse':schemas.FormResponse,
        'AuthRequest':schemas.AuthRequest,
        'ConfirmationRequest':schemas.ConfirmationRequest
    }


class BaseSocketOpts(logger.ConnectionLogger):
    """
    behind the scenes, low level functions to handle the socket operations asynchronoously on the server
    """
    def __init__(self, name, debug=False):
        self.initialize_logger(f"{name} LOGGER")
        self.debug_state = debug

    def debug(self, operation: typing.Literal["SENDING", "RECEIVED", "WAITING"], data: str | typing.Type[schemas.BaseSchema], message: str | None = None):
        if self.debug_state:
            if not isinstance(data, str):
                self.logger.info(f"""{operation}: {data.schema_type}
                                    MESSAGE: {pformat(data.data)}
                                    ERROR: {pformat(data.error)}
                                    MESSAGE CONTENT:{pformat(data.request_content)}""")
            else:
                self.logger.info(data)
            return
        if self.__class__.__name__ == 'ClientSocketOpts':
            return
        elif message:
            self.logger.info(message)


class ClientSocketOpts(BaseSocketOpts):
    """
    synchronous sockets to be used by clients
    """
    def __json_receive_explicit(self, connection):
        json_data = ""
        while True:
            try: 
                json_data = json_data + connection.recv(1024).decode('utf-8') 
                return json.loads(json_data) 
            except ValueError:
                continue
            except BlockingIOError:
                continue

    def __json_send_explicit(self, connection, data):
        json_data = json.dumps(data)
        connection.send(json_data.encode())

    def send(self, data: typing.Type[schemas.BaseSchema]):
        self.debug('SENDING', data)
        self.__json_send_explicit(self.connection, data.dict())

    def receive(self) -> typing.Type[schemas.BaseSchema]:
        self.debug("WAITING", "Awaiting data receive")
        data = self.__json_receive_explicit(self.connection)
        schema_type = schema_types[data['schema_type']]
        self.debug('RECEIVED', schema_type(**data))
        return schema_type(**data)


class ServerSocketOpts(BaseSocketOpts):
    """
    behind the scenes, low level functions to handle the socket operations asynchronoously on the server
    """
    def __init__(self, name, debug=False):
        super().__init__(name, debug=debug)
        self.system = ''
        self.pwd = ''

    async def __json_receive_explicit_async(self):
        json_data = ""
        while True:
            try: 
                byte_buffer = await self.reader.read(n=1024)
                json_data += byte_buffer.decode()
                return json.loads(json_data) 
            except ValueError: # if json is invalid, keep going
                continue
            except BlockingIOError: # this is raised in the event that it tries to receive data when no data available due to non blocking
                continue

    async def send(self, data: typing.Type[schemas.BaseSchema]):
        if data.request_content:
            for key, value in data.request_content.items():
                if not isinstance(value, (str, list, tuple, bool, int, dict, set)) and value != None:
                    data.request_content[key] = value.json()
        self.debug('SENDING', data, message=f'Sending data of type {data.schema_type} to the client')
        data.pwd = self.pwd
        data.system = self.system
        json_data = json.dumps(data.dict())
        self.writer.write(json_data.encode())
        await self.writer.drain()

    async def receive(self):
        self.debug("WAITING", "Awaiting data receive", message='Waiting to receive data from the client on this connection')
        data = await self.__json_receive_explicit_async()
        schema_type = schema_types[data['schema_type']]
        formatted_data = schema_type(**data)
        self.debug('RECEIVED', formatted_data, message=f'Successfully received data of type {formatted_data.schema_type}')
        if formatted_data.error:
            raise exceptions.ClientSideError(formatted_data.error)
        return formatted_data
    



import typing


class TimeoutError(Exception):
    """
    Raise exception when timeout time has been exceeded
    """
    def __init__(self):
        super().__init__("Disconnected due to inactivity. Please login again")


class CommandNotFoundError(Exception):
    """
    Raise exception when a command isnt found
    """
    def __init__(self, command: str):
        self.command = command
        super().__init__(f"Command {command} was not found in the command list. See help")
    

class Shutdown(Exception):
    """
    raise error when a shutdown is initiated
    """
    def __init__(self):
        super().__init__(f"shutdown initiated")


class Exit(Exception):
    """
    raise error when an exit is initiated
    """
    def __init__(self):
        super().__init__(f"exit initiated")
    

class NoConfirmationError(Exception):
    """
    raise error when no confirmation is given for a function that needs confirmation to continue
    """
    def __init__(self, function: typing.Callable):
        super().__init__(f"Confirmation was not given to the function {function.__class__.__name__}")


class InvalidCredentialsReceived(Exception):
    """
    raise error when the provided credentials fail
    """
    def __init__(self, *args):
        super().__init__(*args)


class HelpDoesNotExist(AttributeError):
    """
    raise error when the program tries to extract help information from a method, but help is not found
    """
    def __init__(self, command_name):
        super().__init__(f"The command {command_name} has no help menu.\nMust include a docstring with @help: <help information>")


class UnauthorizedAccessError(Exception):
    """
    raise error when an unauthorized IP tries to connect to the server
    """
    def __init__(self, ip):
        super().__init__(f"DANGER! UNAUTHORIZED IP {ip} TRIED TO CONNECT TO THE SERVER. SOMEONE IS TRYING TO PIGGYBACK ON THE TAPIS APPLICATION, AND ATTEMPTING TO STEAL YOUR DATA.")


class ClientSideError(ConnectionError):
    """
    raise error when problem is on client side
    """
    def __init__(self, error_message):
        super().__init__(f"CLIENT SIDE ERROR: {error_message}")


class NoTenantClient(AttributeError):
    """
    raise error when no client for tenant for auth
    """
    def __init__(self, tenant):
        super().__init__(f"No client found for the tenant {tenant}")


import time
import os 
import sys
import json
import webbrowser
import base64
import traceback


from tapipy.tapis import Tapis
from tapipy import tapis
from federatedTenantAuthAPI.get import get_client_code
import requests
import pyperclip


from socketopts import schemas
from utilities import exceptions
from commands.arguments import argument



class ServerSideAuth:
    def create_token_device_grant(link, device_code, client_id):
        response = requests.post(f"https://{link}/v3/oauth2/tokens", json={"grant_type":"device_code", "device_code":device_code, "client_id":client_id})
        parsed_data = json.loads(response.content.decode())
        return parsed_data['result']['access_token']['access_token'], parsed_data['result']['refresh_token']['refresh_token']
    
    def get_tenant_uris(self):
        if not self.t:
            t = Tapis('https://tacc.tapis.io')
        else:
            t = self.t
        tenants = t.tenants.list_tenants()
        uri_list = list()
        for tenant in tenants:
            uri_list.append(tenant.base_url.split("//")[1])
        return uri_list
    
    async def password_grant(self, link: str, connection):
        start = time.time()
        username_password_request = schemas.AuthRequest(auth_request_type='password',
                                                        request_content={"username":argument.Argument('username', size_limit=(1, 100), arg_type='str_input'), "password":argument.Argument('password', arg_type='secure')},
                                                        message={"message":"enter your TACC username and password"})
        for attempt in range(1, 4):
            await connection.send(username_password_request)
            username_password_response = await connection.receive()
            username = username_password_response.request_content['username']
            password = username_password_response.request_content['password']
            try:
                self.t = Tapis(f"https://{link}",
                               username=username,
                               password=password)
                print(self.t.base_url)
                self.t.get_tokens()
                break
            except exceptions.ClientSideError as e:
                raise e
            except Exception as e:
                if attempt > 3:
                    username_password_request.message = None
                    username_password_request.request_content = None
                    username_password_request.error = "All attempts used, exiting"
                    connection.send(username_password_request)
                    if self.initial:
                        await self.aggregate_command_map['shutdown']
                    raise exceptions.InvalidCredentialsReceived()

                username_password_request.error = f"Authentication failure, attempt {attempt} of 3"

        self.username = username
        self.password = password
        self.url = link
        self.access_token = self.t.access_token

        self.logger.info(f"initiated in {time.time()-start}")

        return f"Successfully initialized tapis service on {self.url}"
    
    async def device_code_grant(self, link: str, connection):
        start = time.time()
        self.t = Tapis(f"https://{link}", resource_set='dev')
        client_id = get_client_code(link)
        if not client_id:
            raise exceptions.NoTenantClient(link)
        authentication_information = self.t.authenticator.generate_device_code(client_id=client_id)
        payload = schemas.AuthRequest(auth_request_type='device_code',
                                      message={"message":"Go to the URL if it doesnt open automatically and enter the user code to authenticate. Then click y when you enter the code\nUser code copied to clipboard", 
                                               "url":authentication_information.verification_uri, 
                                               "user_code": authentication_information.user_code},
                                      request_content={'Entered User Code':argument.Argument('Entered User Code', arg_type='confirmation')})
        await connection.send(payload)
        pyperclip.copy(authentication_information.user_code)
        webbrowser.open(authentication_information.verification_uri)
        confirmation = await connection.receive()
        if not confirmation.request_content['Entered User Code']:
            raise RuntimeError(f"User indicated that device code auth failed") # get rid of this feature and switch to polling
        start_time = time.time()
        while True:
            try:
                access_info = self.t.authenticator.create_token(grant_type="device_code", device_code=authentication_information.device_code, client_id=client_id)
                break
            except Exception as e:
                print(e)
                time.sleep(1)
                if time.time() - start_time > 60:
                    raise RuntimeError("Timeout while polling for authenticator token")

        self.t = Tapis(f"https://{link}",
                       access_token=access_info.access_token.access_token,
                       refresh_token=access_info.refresh_token.refresh_token,
                       resource_set="dev")
        
        self.username = self.t.authenticator.get_userinfo().username
        self.url = link
        self.access_token = self.t.access_token

        self.logger.info(f"initiated in {time.time()-start}")

        return f"Successfully initialized tapis service on {self.url}"

    async def federated_grant(self, link, connection):
        start = time.time()
        auth_link = rf"https://{link}/v3/oauth2/webapp"
        payload = schemas.AuthRequest(auth_request_type='federated',
                                      message={
                                          "message":"Go to the URL if it doesnt open and enter your account information. Then enter the user code to this application",
                                          "url":auth_link
                                            }, 
                                      request_content={
                                          "access_token":argument.Argument('access_token', arg_type='str_input')
                                          })
        await connection.send(payload)
        webbrowser.open(auth_link)
        response_code_message: schemas.AuthRequest = await connection.receive()
        
        self.t = Tapis(f"https://{link}",
                       access_token=response_code_message.request_content['access_token'].strip())

        self.username = self.t.authenticator.get_userinfo().username
        self.url = link
        self.access_token = self.t.access_token

        self.logger.info(f"initiated in {time.time()-start}")

        return f"Successfully initialized tapis service on {self.url}"

    async def auth_startup(self, connection):
        session_auth_type_request = schemas.AuthRequest(request_content={"Tenant URI":argument.Argument('Tenant URI', arg_type='str_input', choices=self.get_tenant_uris()), "auth_type":argument.Argument('auth_type', choices=['password', 'device_code', 'federated'], arg_type='str_input')},
                                                        auth_request_type="requested",
                                                        message={"message":"Enter the URI of the Tapis tenant you wish to connect to, then select your auth type from the options below",
                                                                 "grant options":['password', 'device_code', 'federated']})
        while True:
            try:
                await connection.send(session_auth_type_request)
                session_auth_type_response: schemas.FormResponse = await connection.receive()
                auth_type = session_auth_type_response.request_content['auth_type']
                if auth_type not in ('password', 'device_code', 'federated'):
                    raise exceptions.InvalidCredentialsReceived()
                self.auth_type = auth_type
                link = session_auth_type_response.request_content['Tenant URI']
                self.t = Tapis(f"https://{link}")#, resource_set='dev')
                break
            except exceptions.InvalidCredentialsReceived:
                session_auth_type_request.error = "Invalid auth type received"
            except tapis.errors.BaseTapyException:
                session_auth_type_request.error = "Invalid tenant URI received, try again"

        if auth_type != "password":
            session_password_request = schemas.AuthRequest(auth_request_type=auth_type, request_content={"password":argument.Argument('password', arg_type='secure', size_limit=(6, 50))},
                                                            message={"message":"Please create a session password. The password must be at least 6 characters long"})
            while True:
                await connection.send(session_password_request)
                session_password_response: schemas.FormResponse = await connection.receive()
                requested_password = session_password_response.request_content['password']
                if len(requested_password) >= 6:
                    self.password = requested_password
                    break
                session_password_request.error = "The password provided was too short, the password must be at least 6 characters!!!"
        
        while True:
            try:
                if auth_type == "federated":
                    await self.federated_grant(link, connection)
                elif auth_type == "device_code":
                    connection.set_status_device_authenticating()
                    await self.device_code_grant(link, connection)
                    connection.set_status_closed()
                else:
                    await self.password_grant(link, connection)
                break
            except (exceptions.ClientSideError, exceptions.InvalidCredentialsReceived, exceptions.NoTenantClient, RuntimeError) as e:
                raise e
            except Exception as e:
                error_str = traceback.format_exc()
                print(error_str)
                continue
        success_message = schemas.AuthRequest(auth_request_type="success", message={
            "message":f"Successfully authenticated with {auth_type} authentication",
            "username":self.username,
            "url":link})
        self.configure_decorators(self.username, self.password)
        self.update_credentials(self.t, self.username, self.password)
        await connection.send(success_message)
            




class CHECK_EXPLICIT_ID:
    def __init__(self, arg_name: str):
        self.arg_name = arg_name
    
    def __call__(self, kwargs):
        if self.arg_name not in kwargs or not kwargs[self.arg_name]:
            kwargs[self.arg_name] = kwargs['connection'].system
        return kwargs


class CHECK_PWD:
    """
    support the invocation of relative paths for tapis systems
    """
    def __init__(self, dir_simplify_args: tuple):
        self.dir_simplify_args = dir_simplify_args

    def __go_back_checker(self, index: int, path_list: list):
        back_count = 0
        for element in path_list[index:]:
            if element != "..":
                break
            back_count += 1
        return back_count

    def __simplify_path(self, path: list):
        index = 0
        length = len(path)
        try:
            while index < length:
                if path[index] == ".":
                    path.pop(index)
                    continue
                elif path[index] == "..":
                    back_count = self.__go_back_checker(index, path)
                    desired_len = len(path) - (2 * back_count)
                    while len(path) != desired_len:
                        path.pop(index-back_count)
                    continue
                index += 1
        except IndexError:
            pass
        finally:
            path = "/".join(path)
            if not path:
                path = "/"
        return path
    
    def __relative_to_absolute(self, absolute_path: str, relative_path: str):
        if not relative_path:
            return absolute_path
        elif absolute_path[-1] == "/":
            return f"{absolute_path}{relative_path}"
        return f"{absolute_path}/{relative_path}"

    def __call__(self, kwargs):
        for file_argument_name in self.dir_simplify_args:
            if not kwargs[file_argument_name] or kwargs['connection'].pwd not in kwargs[file_argument_name]:
                file = self.__relative_to_absolute(kwargs['connection'].pwd, kwargs[file_argument_name])
                kwargs[file_argument_name] = self.__simplify_path(file.split("/"))
        return kwargs


if __name__ != "__main__":
    from . import baseCommand, decorators
    from .arguments import argument
    from .Apps import appCommands
    from commands import commandOpts
    Argument = argument.Argument


class hide_job(baseCommand.BaseCommand):
    """
    @help: hide a job if its already completed
    """
    required_arguments = [
        Argument('jobUuid', positional=True)
    ]
    async def run(self, *args, **kwargs):
        return self.t.jobs.hideJob(**kwargs)
    

class unhide_job(hide_job):
    """
    @help: reveal a job previously hidden
    """
    async def run(self, *args, **kwargs):
        return self.t.jobs.unhideJob(**kwargs)
    

class cancel_job(hide_job):
    """
    @help: cancel a job as it executes
    """
    required_arguments = [
        Argument('jobUuid', positional=True),
        Argument('confirm', arg_type='confirmation')
    ]
    async def run(self, *args, **kwargs):
        return self.t.jobs.cancelJob(**kwargs)
    

class get_job(hide_job):
    """
    @help: get a job and its information
    """
    return_fileds = ['appId', 'appVersion', 'execSystemId', 'uuid']
    async def run(self, *args, **kwargs):
        return self.t.jobs.getJob(**kwargs)
    

class get_job_history(hide_job):
    """
    @help: display the history of changes made to a job
    """
    return_fileds = ['event', 'created']
    optional_arguments = [
        Argument('limit', data_type='int'),
        Argument('skip', data_type='int'),
    ]
    async def run(self, *args, **kwargs):
        return self.t.jobs.getJobHistory(**kwargs)
    

class get_jobs(baseCommand.BaseCommand):
    """
    @help: list all available jobs on the system
    """
    return_fileds = ['appId', 'appVersion', 'execSystemId', 'uuid']
    async def run(self, *args, **kwargs):
        return self.t.jobs.getJobList()
    

class download_job_output(baseCommand.BaseCommand):
    """
    @help: download the job output to an output path
    """
    required_arguments = [
        Argument('jobUuid'),
        Argument('outputPath')
    ]
    optional_arguments = [
        Argument('compress', action='store_true'),
        Argument('format', description='something like zip or tar')
    ]
    async def run(self, *args, **kwargs):
        self.t.jobs.getJobOutputDownload(**kwargs)
        return f"job {kwargs['jobUuid']} output successfully downloaded"
    

class get_job_status(baseCommand.BaseCommand):
    """
    @help: retrieve the status of a job
    """
    required_arguments = [
        Argument('jobUuid')
    ]
    async def run(self, *args, **kwargs):
        return self.t.jobs.getJobStatus(**kwargs)
    

class get_job_last_message(baseCommand.BaseCommand):
    """
    @help: retrieve the last message of a job
    """
    required_arguments = [
        Argument('jobUuid')
    ]
    async def run(self, *args, **kwargs):
        return self.t.jobs.getJob(**kwargs).lastMessage

class resubmit_job(baseCommand.BaseCommand):
    """
    @help: resubmit a job with the same arguments as the original submission
    """
    return_fields = ['name', 'status', 'uuid', 'appId', 'execSystemId']
    required_arguments = [
        Argument('jobUuid'),
    ]
    async def run(self, *args, **kwargs):
        return self.t.jobs.resubmitJob(**kwargs)
    

class submit_job(appCommands.assign_default_job_attributes):
    """
    @help: submit a job to be run on the select system based on a pre-existing app
    """
    return_fields = ['name', 'status', 'uuid', 'appId', 'execSystemId']
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('execSystemId')]
    required_arguments = [
        Argument('appId', size_limit=(1, 80), positional=True),
        Argument('name', positional=True)
    ]
    async def run(self, *args, **kwargs):
        kwargs = appCommands.get_latest_version(self.t, kwargs)
        return self.t.jobs.submitJob(**kwargs)
    

class share_job(baseCommand.BaseCommand):
    """
    @help: share a job with the select grantee user
    """
    required_arguments = [
        Argument('jobUuid')
    ]
    optional_arguments = [
        Argument('grantee', description='Who are you sharing this with?'),
        Argument('jobResource', choices=['JOB_HISTORY', 'JOB_RESUBMIT_REQUEST', 'JOB_OUTPUT', 'JOB_INPUT']),
        Argument('jobPermission', choices=['READ'])
    ]
    async def run(self, *args, **kwargs):
        return self.t.jobs.shareJob(**kwargs)
    

class get_job_share(baseCommand.BaseCommand):
    """
    @help: get sharing information on a specific job
    """
    return_fields = ['jobUuid', 'jobResource', 'grantee']
    required_arguments = [
        Argument('jobUuid')
    ]
    async def run(self, *args, **kwargs):
        return self.t.jobs.getJobShare(**kwargs)
    

class delete_job_share(baseCommand.BaseCommand):
    """
    @help: delete a job sharing
    """
    required_arguments = [
        Argument('jobUuid'),
        Argument('grantee')
    ]
    async def run(self, *args, **kwargs):
        kwargs['user'] = kwargs['grantee']
        kwargs.pop('grantee')
        return self.t.jobs.deleteJobShare(**kwargs)


class subscribe_to_job(baseCommand.BaseCommand):
    """
    @help: subscribe to a job in order to receive notifications on its status by email or such
    """
    required_arguments = [
        Argument('jobUuid'),
        Argument('eventCategoryFilter', choices=[
            'JOB_NEW_STATUS',
            'JOB_INPUT_TRANSACTION_ID',
            'JOB_ARCHIVE_TRANSACTION_ID',
            'JOB_SUBSCRIPTION',
            'JOB_SHARE_EVENT',
            'JOB_ERROR_MESSAGE',
            'JOB_USER_EVENT',
            'ALL'
            ]),
        Argument('deliveryTargets', arg_type='input_list', data_type=argument.Form('deliveryTarget', required_arguments=[
            Argument('deliveryMethod', choices=['WEBHOOK', 'EMAIL', 'QUEUE', 'ACTOR']),
            Argument('deliveryAddress')
        ]))
    ]
    optional_arguments = [
        Argument('description', arg_type='str_input'),
    ]
    async def run(self, *args, **kwargs):
        self.t.jobs.subscribe(**kwargs)
        return f"successfully subscribed to {kwargs['eventCategoryFilter']} event(s) at the selected addresses"
    

class get_subscriptions(baseCommand.BaseCommand):
    """
    @help: list subscriptions
    """
    return_fields = ['name', 'uuid', 'subjectFilter']
    required_arguments = [
        Argument('jobUuid')
    ]
    async def run(self, *args, **kwargs):
        return self.t.jobs.getSubscriptions(**kwargs)
    

class delete_subscriptions(baseCommand.BaseCommand):
    """
    @help: delete a subscription if you no longer want to receive its notifications
    """
    required_arguments = [
        Argument('confirm', arg_type='confirmation')
    ]
    optional_arguments = [
        Argument('jobUuid'),
        Argument('subscriptionUuid', mutually_exclusive_with=['jobUuid'])
    ]
    async def run(self, *args, **kwargs):
        if kwargs['jobUuid']:
            uuid = kwargs['jobUuid']
        elif kwargs['subscriptionUuid']:
            uuid = kwargs['subscriptionUuid']
        else:
            raise ValueError('Must either specify jobUuid or subscriptionUuid')
        self.t.jobs.deleteSubscriptions(uuid=uuid)
        return f"subscription {kwargs['uuid']} deleted"